/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.component.framework.util;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.messaging.Fault;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;

import org.apache.axiom.om.impl.llom.util.AXIOMUtil;
import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 * This class allows to handle a JBI Fault as a SOAP Fault
 *
 * @author nsalatge - EBM Websourcing
 * @since 4.0
 *
 */
public class SOAPFault {


    private Fault jbiFault;

    private List<Element> soapFaults;
    
    /**
     * Soap envelope constant
     * @author Mathieu CARROLLE
     *
     */
    public enum SoapEnvelopeConstants {
        SOAP11_NS("http://schemas.xmlsoap.org/soap/envelope/"),
        SOAP12_NS("http://www.w3.org/2003/05/soap-envelope");
        
        /**
         * 
         * @param pattern
         * @return
         */
        public static SoapEnvelopeConstants valueOf(final URI pattern) {
            SoapEnvelopeConstants result = null;
            if (pattern != null) {
                for (SoapEnvelopeConstants soap : SoapEnvelopeConstants.values()) {
                    if (soap.nameSpace.equals(pattern.toString())) {
                        result = soap;
                    }
                }
            }
            return result;
        }
        
        private final String nameSpace;
        private final URI soapURI;
        
        private SoapEnvelopeConstants(String nameSpace) {
            this.nameSpace = nameSpace;
            try {
                this.soapURI = new URI(nameSpace);
            } catch (final URISyntaxException e) {
                throw new Error("Unexpected Error in URI namespace syntax", e); // NOPMD
            }
        }
        
        /**
         * 
         * @return
         */
        public URI value() {
            return this.soapURI;
        }

        /**
         * Please use this equals method instead of using :<code>
         * value().equals(mep)
         * </code>
         * which is almost 10 times slower...
         * 
         * @param mep
         * @return
         */
        public boolean equals(URI mep) {
            return toString().equals(mep.toString());
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.lang.Enum#toString()
         */
        public String toString() {
            return this.nameSpace;
        }
    }
    
    /**
     * SOAP Version Constants for the Message Exchange Patterns.
     * 
     */
    public enum BindingSoapVersionConstants {
        SOAP11_NS("http://schemas.xmlsoap.org/wsdl/soap/"),
        SOAP12_NS("http://schemas.xmlsoap.org/wsdl/soap12/");
        /**
         * 
         * @param pattern
         * @return
         */
        public static BindingSoapVersionConstants valueOf(final URI pattern) {
            BindingSoapVersionConstants result = null;
            if (pattern != null) {
                for (BindingSoapVersionConstants soap : BindingSoapVersionConstants.values()) {
                    if (soap.nameSpace.equals(pattern.toString())) {
                        result = soap;
                    }
                }
            }
            return result;
        }

        private final String nameSpace;
        private final URI soapURI;

        /**
         * Creates a new instance of {@link MEPConstants}
         * 
         * @param nameSpace
         */
        private BindingSoapVersionConstants(final String nameSpace) {
            this.nameSpace = nameSpace;
            try {
                this.soapURI = new URI(nameSpace);
            } catch (final URISyntaxException e) {
                throw new Error("Unexpected Error in URI namespace syntax", e); // NOPMD
            }
        }

        /**
         * 
         * @return
         */
        public URI value() {
            return this.soapURI;
        }

        /**
         * Please use this equals method instead of using :<code>
         * value().equals(mep)
         * </code>
         * which is almost 10 times slower...
         * 
         * @param mep
         * @return
         */
        public boolean equals(URI mep) {
            return toString().equals(mep.toString());
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.lang.Enum#toString()
         */
        public String toString() {
            return this.nameSpace;
        }
    }


    /**
     * SOAP Fault Code Constants for the Message Exchange Patterns.
     * 
     */
    public enum SoapFaultCodeConstants {
        VERSION_MISMATCH("VersionMismatch"),
        MUST_UNDERSTAND("MustUnderstand"),
        DATA_ENCODING_UNKNOWN("DataEncodingUnknown"),
        SENDER("Sender"),
        RECEIVER("Receiver"),
        CLIENT("Client"),
        SERVER("Server");



        private final String value;

        /**
         * Creates a new instance of {@link FeatureConstants}
         * 
         * @param value
         */
        private SoapFaultCodeConstants(final String value) {
            this.value = value;
        }

        /**
         * 
         * @return
         */
        public String value() {
            return value;
        }

        /**
         * Please use this equals method instead of using :<code>
         * value().equals(value)
         * </code>
         * which is almost 10 times slower...
         * 
         * @param value
         * @return
         */
        public boolean equals(String value) {
            return toString().equals(value.toString());
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.lang.Enum#toString()
         */
        public String toString() {
            return this.value;
        }
    }

    public SOAPFault(Fault jbiFault) throws PEtALSCDKException {
        this.jbiFault = jbiFault;
        this.soapFaults = new ArrayList<Element>();

        Document jbi = SourceHelper.createDocument(this.jbiFault.getContent());

        NodeList details = null;
        if(jbi.getDocumentElement().getNamespaceURI().equals(SoapEnvelopeConstants.SOAP11_NS.toString())) {
            details = jbi.getDocumentElement().getElementsByTagName("detail");
        } else if(jbi.getDocumentElement().getNamespaceURI().equals(SoapEnvelopeConstants.SOAP12_NS.toString())) {
            details = jbi.getDocumentElement().getElementsByTagName("Detail");
        }

        if((details != null)&&(details.getLength() > 0)) {
            for (int i = 0; i < details.getLength(); i++) {
                if(details.item(i).getNodeType() == Node.ELEMENT_NODE) {
                    soapFaults.add((Element) details.item(i));
                }
            }
        }
    }

    public List<Element> getSoapFaults() {
        return soapFaults;
    }

    public static Source createSOAP11Fault(SoapFaultCodeConstants code, String reason, List<Element> details) throws PEtALSCDKException {
        return createFault(BindingSoapVersionConstants.SOAP11_NS, code, reason, details);
    }
    
    public static Source createSOAP12Fault(SoapFaultCodeConstants code, String reason, List<Element> details) throws PEtALSCDKException {
        return createFault(BindingSoapVersionConstants.SOAP12_NS, code, reason, details);
    }
    
    
    private static Source createFault(BindingSoapVersionConstants version, SoapFaultCodeConstants code, String reason, List<Element> details) throws PEtALSCDKException {
        Source res = null;
        org.apache.axiom.soap.SOAPEnvelope env = null;

        String ns = null;
        SoapFaultCodeConstants correctCode = code; 
        if(version.equals(BindingSoapVersionConstants.SOAP11_NS)) {
            env = new org.apache.axiom.soap.impl.llom.soap11.SOAP11Factory().getDefaultFaultEnvelope();
            ns = "{" + BindingSoapVersionConstants.SOAP11_NS + "}";
            if(code.equals(SoapFaultCodeConstants.DATA_ENCODING_UNKNOWN)) {
                throw new PEtALSCDKException("Fault code unknown {" + code.value + "} in this soap version: " + version);
            }
            if(code.equals(SoapFaultCodeConstants.RECEIVER)) {
                correctCode = SoapFaultCodeConstants.CLIENT;
            } else if(code.equals(SoapFaultCodeConstants.SENDER)) {
                correctCode = SoapFaultCodeConstants.SERVER;
            } 
        } else if(version.equals(BindingSoapVersionConstants.SOAP12_NS)) {
            env = new org.apache.axiom.soap.impl.llom.soap12.SOAP12Factory().getDefaultFaultEnvelope();
            ns = "{" + BindingSoapVersionConstants.SOAP12_NS + "}";
            if(code.equals(SoapFaultCodeConstants.CLIENT)) {
                correctCode = SoapFaultCodeConstants.RECEIVER;
            } else if(code.equals(SoapFaultCodeConstants.SERVER)) {
                correctCode = SoapFaultCodeConstants.SENDER;
            } 
        } else {
            throw new PEtALSCDKException("Fault unknown");
        }

        org.apache.axiom.soap.SOAPFault fault = env.getBody().getFault();

        // create fault code
        org.apache.axiom.soap.SOAPFaultCode faultCode = fault.getCode();
        if(version.equals(BindingSoapVersionConstants.SOAP12_NS)) {
            faultCode.getValue().setText(QName.valueOf(ns + correctCode.value));
            fault.setCode(faultCode);
        } else {   
            faultCode.setText(QName.valueOf(ns + correctCode.value));
        }

        // create reason
        org.apache.axiom.soap.SOAPFaultReason faultReason = fault.getReason();
        if(version.equals(BindingSoapVersionConstants.SOAP12_NS)) {
            faultReason.getSOAPFaultText(null).setLang("en");
            faultReason.getSOAPFaultText(null).setText(reason);
        } else {
            faultReason.setText(reason);
        }
        
        // create details
        org.apache.axiom.soap.SOAPFaultDetail faultDetail = fault.getDetail();
        try {
            for (Element element : details) {
                faultDetail.addDetailEntry(AXIOMUtil.stringToOM(XMLUtil.createStringFromDOMNode(element)));
            }
            res = AxiomUtils.createSource(fault);
        } catch (XMLStreamException e) {
            throw new PEtALSCDKException("Error: Impossible to detils in soap fault", e);
        } catch (TransformerException e) {
            throw new PEtALSCDKException("Error: Impossible to detils in soap fault", e);
        } catch (IOException e) {
            throw new PEtALSCDKException("Error: Impossible to detils in soap fault", e);
        }

        return res;
    }

}
